#
# -*- coding: utf-8 -*-
#
# Suborder_Printer.py 
#
# Watches a configured IMAP folder for Mails and prints
# them on the given printer.
#
# Prequesits:
#
# - To print German umlauts (äöüÄÖÜß) correctly, configure
#   the EPSON printer to Int+ernational Chareacter Set "Germany".
#   -> Set Printer Off, and turn On while pressing [Feed]
#   -> Mode Selection
#   -> 3: Customize Value Settings
#   -> 8: Default Character
#   -> 2: International Character Set
#   -> 1: The Americas, Europe
#   -> 3: Germany
#
# - The printer name is hard coded as "EPSON TM-T88V Receipt"
#
# - Coding of printer data is "Latin-1"
#
# - In an multi-store config, the format of the workstation
#   name has to by like:
#
#    	XXXXX-12345-9-9
#   or
#       XXXXXXXX-12345
#
#   where 12345 is the store number, and XXXXX letters and
#   9 are digits. The first format is SubwayPOS format, the
#   second is for Subshop/2000.
#
#   Then make IMAP folders with the store numbers, and use
#   rules in your IMAP account to sort the mails into these
#   IMAP folders. 
#
# Copyright (C) 2020 SUBWAY Pegnitz, Daniel Groß
#
history = "22.03.2020 - dg - modified for release"
#--------------------------------------------------------------------------
# CONFIGURATION SECTION. ;-)
#
# e-mail server & account
IMAP_SERVER   = 'imap.1und1.de'                 # <- IMAP mail server
EMAIL_ACCOUNT = 'your_email_account_login'      # <- IMAP mail account
PASSWORD      = 'your_email_account_password'   # <- IMAP mail password

# the printer
PRINTER       = 'EPSON TM-T88V Receipt'         # <- Receipt printer

# SUBWAY store number
# => IMAP folder name with new orders
STORE         = 'your_store_number'             # <- IMAP store folder

# IMAP archive folder
ARCHIVE = 'Archive'                             # <- IMAP archive folder
#--------------------------------------------------------------------------

import os
import re
import sys
import datetime
import email
import imaplib
import mailbox
import time
import platform
import traceback
import smtplib

try:
    import win32print
except:
    print("Not Windows")

from email.header import Header, decode_header, make_header
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email.encoders import encode_base64
from email.encoders import encode_quopri

#--------------------------------------------------------------------------
# display message
if STORE = 'your_store_number':
    store_match = re.search(r'.*-(\d+)-\d-\d$', platform.node())
    if store_match is not None:
        STORE = store_match.group(1)
    else:
        store_match = re.search(r'\d+$', platform.node())
        if store_match is not None:
            STORE = store_match.group()


# CR/LF or LF only?
if ('idlelib' not in sys.modules):
    LINE_END = '\r\n'
else:
    LINE_END = '\n'

# EPSON TM-T88V printer header and footer (allways CR/LF)
# 42 cols  [....+....1....+....2....+....3....+....4..]
HEADER   = LINE_END + '\x1bM \x1b3' + LINE_END + ' ' + LINE_END \
         + '******************************************' + LINE_END \
         + '*********  SUBORDER  BESTELLUNG  *********' + LINE_END
FOOTER   = LINE_END \
         + '*********  SUBORDER  BESTELLUNG  *********' + LINE_END \
         + '******************************************' + LINE_END \
         + ' ' + LINE_END \
         + ' ' + LINE_END \
         + ' ' + LINE_END \
         + ' ' + LINE_END \
         + ' ' + LINE_END \
         + '\x1dV0' + LINE_END + '\f' + LINE_END
# ESC seq to cut on EPSON TM-T88V

#--------------------------------------------------------------------------
# display message
def log(message):

    print(time.strftime("%H:%M:%S", time.localtime()) +": " + message)
    return

#--------------------------------------------------------------------------
# print to EPSON printer
def print_order_slip(text):


    #if ('idlelib' not in sys.modules):
    if (STORE != 'TEST'):
        print(text)
        raw_data = bytes(HEADER + text + FOOTER, 'cp1252')
        hPrinter = win32print.OpenPrinter(PRINTER)
        try:
            hJob = win32print.StartDocPrinter(hPrinter, 1, ("SUBORDER", None, "RAW"))
            try:
                win32print.StartPagePrinter(hPrinter)
                win32print.WritePrinter(hPrinter, raw_data)
                win32print.EndPagePrinter(hPrinter)
            finally:
                win32print.EndDocPrinter(hPrinter)
        finally:
            win32print.ClosePrinter(hPrinter)

        #with open(PRINT_FILE, mode='w', encoding='cp858') as lpt:
        #    lpt.write(HEADER)
        #    lpt.write(text)
        #    lpt.write(FOOTER)
    else:
        # debug in IDLE
        print(text, end='')

#--------------------------------------------------------------------------
# Debugging by E-Mail
def debug_mail(text):

    # create MIME message
    msg = MIMEMultipart()
    msg['Subject'] = "SUBORDER Debug"
    msg['From'] = "SUBORDER <suborder@subway51704.eu>"
    msg['To'] = "daniel.gross@subway-pegnitz.de"
    html = """\
<html>
<head></head>
<body>
<pre style="font-size: 14px">
""" + text + """\
</pre>
</body>
</html>
"""           
    # set message text and encode (ESC chars)
    text = MIMEText(html, 'html', 'utf-8')
    text.add_header("Content-Disposition", "inline")
    msg.attach(text)
    # send via SMTP/SSL
    s = smtplib.SMTP_SSL()
    #s.set_debuglevel(1)   # activate if needed
    s.connect("smtp.1und1.de")
    s.login("suborder@subway51704.eu", "nv?Jks4l")
    s.sendmail(msg['From'], msg['To'], msg.as_string())
    s.quit()

#--------------------------------------------------------------------------
# analyze and print order
def process_order(body_text):

    top_lines = ''
    sub_lines = ''
    where_am_i = '*TOP'
    for line in body_text.splitlines():

        # some "shortening" to better fit the 42-col EPSON printer
        line = line.replace('Gewünschte Abholzeit:', 'Gew. Abholzeit:')
        if line == '-'*66:
            line = '-'*42
        if re.search(r'\(\d+\) \d+ x Sandwich', line) is not None:
            line = line.replace('x Sandwich', 'x')
        if ' EUR*' in line:
            line = line.replace(' EUR*', '€')

        # blank/empty lines must have 1 space - otherwise compressed
        if line.strip() == '':
            line = ' '
        
        # FreshToasted? really? and an empty line please
        if 'FreshToasted:' in line:
            line = line.replace('FreshToasted:', 'getoastet:') + LINE_END + ' '

        # Sosse - empty line before
        if '- Sosse:' in line:
            line = ' ' + LINE_END + line

        # Wünsche - empty line before
        if '- Wünsche!:' in line:
            line = ' ' + LINE_END + line

        # Käse / Bacon / Double Meat -> before getoastet
        if 'KÄSE:' in line.upper() \
        or 'BACON?:' in line.upper():
            idx = sub_lines.find(' - getoastet:')
            if idx > 0 and where_am_i == '*SUB':
                sub_lines = sub_lines[:idx] \
                          + line + LINE_END \
                          + sub_lines[idx:]
                continue
        if 'DOPPELT FLEISCH?:' in line.upper():
            idx = sub_lines.find(' - Größe:')
            if idx > 0:
                idx = sub_lines.find(LINE_END, idx)
                if idx > 0 and where_am_i == '*SUB':
                    sub_lines = sub_lines[:idx] \
                              + LINE_END + line \
                              + sub_lines[idx:]
                    continue

        # resolv party/sub platters
        if 'X SUB-PLATTE -' in line.upper():
            if 'DIE KLASSISCHE' in line.upper():
                line = line + LINE_END \
                     + '         - Italian B.M.T.' + LINE_END \
                     + '         - Chicken Breast' + LINE_END \
                     + '         - Tuna          ' + LINE_END \
                     + '         - Veggie DeLITE ' + LINE_END \
                     + ' ' + LINE_END
            if 'DIE GENUSSVOLLE' in line.upper():
                line = line + LINE_END \
                     + '         - Italian B.M.T.' + LINE_END \
                     + '         - Chicken Fajita' + LINE_END \
                     + '         - Ham           ' + LINE_END \
                     + '         - Chicken Breast' + LINE_END \
                     + ' ' + LINE_END
            if 'DIE AKTIVE' in line.upper():
                line = line + LINE_END \
                     + '         - Turkey & Ham  ' + LINE_END \
                     + '         - Turkey        ' + LINE_END \
                     + '         - Veggie DeLITE ' + LINE_END \
                     + '         - Chicken Breast' + LINE_END \
                     + ' ' + LINE_END
            if 'DIE GRÜNE' in line.upper():
                line = line + LINE_END \
                     + '         - Veggie Patty  ' + LINE_END \
                     + '         - Veggie Patty  ' + LINE_END \
                     + '         - Veggie DeLITE ' + LINE_END \
                     + '         - Veggie DeLITE ' + LINE_END \
                     + ' ' + LINE_END

        # find lines like "(1) 1 x Sandwich ...."
        if re.search(r'\(\d+\) \d+ x', line) is not None:
            where_am_i = '*SUB'
            if sub_lines != '':
                print_order_slip(top_lines + sub_lines)
            sub_lines = ''

        if where_am_i == '*TOP':
            top_lines = top_lines + line + LINE_END
        else:
            sub_lines = sub_lines + line + LINE_END

    if sub_lines != '':
        if '*DEBUG*' not in sub_lines: 
            # print last sub
            print_order_slip(top_lines + sub_lines)
        else:
            debug_mail(top_lines + sub_lines)

#--------------------------------------------------------------------------
# check mailbox
def check_mailbox():

    # connect and login to server
    M = imaplib.IMAP4_SSL(IMAP_SERVER)
    try:
        M.login(EMAIL_ACCOUNT, PASSWORD)
    except imaplib.IMAP4.error:
        log("LOGIN FAILED!!! ")
        return

    # select correct mailbox
    #print(M.list())
    rv, data = M.select(STORE)
    if rv != 'OK':
        M.list()
        log("SELECT MAILBOX FAILED!!! ")
        return

    # search mail messages in mailbox        
    rv, data = M.search(None, "ALL")
    if rv != 'OK':
        log("No messages found!")
        return
    
    # give time to fetch mail
    time.sleep(0.5)

    # fetch messages one-by-one and process
    #print(data[0].split())
    for num in data[0].split():
        rv, data = M.fetch(num, '(RFC822)')
        if rv != 'OK':
            log("ERROR getting message", num)
            return

        # decode byte-stream to python object
        if data is not None:
            msg = email.message_from_bytes(data[0][1])
            subject = '%s' % make_header(decode_header(msg['Subject']))

            # walk thru MIME parts and find plain text body
            body_text = ''
            for part in msg.walk():
                if part.get_content_type() == 'text/plain':
                    body_text = part.get_payload(decode=True).decode('utf-8')

            if body_text != '':
                # body text found
                log(subject)

                # analyze and print the order
                process_order(LINE_END + body_text)

                # if not debugging - move mail to archive
                if ('idlelib' not in sys.modules) and (STORE != 'TEST'):
                    rc = M.copy(num, ARCHIVE)
                    if rc[0] == 'OK':
                        M.store(num, '+FLAGS', '\\Deleted')
                        M.expunge()
                        time.sleep(3.0) # give time to expunge
                # we found a body text - that's it - next message
                continue
    # close IMAP connection
    M.close()
    M.logout()
    return

#--------------------------------------------------------------------------
# Main script

# Script modification time at start-up
SCRIPT_MTIME = os.path.getmtime(os.path.abspath(__file__))

print("")
#      ....+....1....+....2....+....3....+....4....+....5....+....6....+....7....+....8
print("#------------------------------------------------------------------------------")
print("# Script   : "+os.path.abspath(__file__))
print("# Version  : "+history)
print("# Release  : "+time.strftime("%c", time.localtime(SCRIPT_MTIME)))
print("#")
print("# Watching : "+EMAIL_ACCOUNT+"  IMAP folder : /"+STORE)
print("# Printer  : "+PRINTER)
print("#------------------------------------------------------------------------------")

try:
    # endless loop (needs less than 30 seconds on a Cray ;-)
    while True:

        # check for script update - spawn self und exit
        if SCRIPT_MTIME != os.path.getmtime(os.path.abspath(__file__)):
            log('Script has been modified - re-starting.')
            os.execv(sys.executable, [sys.executable] + sys.argv)
            exit(0) # <- never reached - but paranoia

        # check mailbox and process messages
        check_mailbox()
                    
        # wait some seconds for the next check
        # if running in IDLE -> debug - don't sleep, just exit
        if ('idlelib' in sys.modules) or (STORE == 'TEST'):
            log('Running in IDLE - exit.')
            break
        else:
            print(time.strftime("%H:%M:%S ...", time.localtime()), end='\r')
            time.sleep(10)

except:
    #      ....+....1....+....2....+....3....+....4....+....5....+....6....+....7....+....8
    print(' ')
    print('*'*79)
    print(time.strftime("%c", time.localtime()))
    print('Exception occured ... waiting for 30 seconds and trying to restart ...')
    print(sys.exc_info()[0])
    print(sys.exc_info()[1])
    print('Traceback (most recent call last):')
    print(traceback.print_tb(sys.exc_info()[2]))
    print('*'*79)
    print()
    #if ('idlelib' not in sys.modules):
    time.sleep(30)
    os.execv(sys.executable, [sys.executable] + sys.argv)
    exit(0) # <- never reached - but paranoia
    
# ==================================== end-of-script ====================================
